
#include "../include/sys/socket.h"
#include "../include/netinet/in.h"
#include <sys/fsext.h>
#include <errno.h>
#include <libc/dosio.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define SOCKERR_INVALID_PARAM		0x00000100
#define SOCKERR_NO_MEMORY		0x00000200
#define SOCKERR_INVALID_SOCKET		0x00000300
#define SOCKERR_ALREADY_BOUND		0x00000400
#define SOCKERR_NOT_BOUND		0x00000500
#define SOCKERR_ACCESS			0x00000600
#define SOCKERR_INTERNAL		0x00000700
#define SOCKERR_FD_INUSE		0x00000800
#define SOCKERR_INFINITE_WAIT		0x00000900
#define SOCKERR_NOT_CONNECTED		0x00000a00
#define SOCKERR_WOULD_BLOCK		0x00000b00
#define SOCKERR_NOT_LISTENING		0x00000c00

int __quiet_socket = 0;

int
converttoerrno (int res)
{
  res = -res;
  if (!__quiet_socket)
      printf ("socket result %d\n", res);
  switch (res) {
  case SOCKERR_INVALID_PARAM:
    return EINVAL;

  case SOCKERR_NO_MEMORY:
    return ENOMEM;

  case SOCKERR_INVALID_SOCKET:
    return EBADF;

  case SOCKERR_ALREADY_BOUND:
    return EINVAL;

  case SOCKERR_NOT_BOUND:
    return EINVAL;

  case SOCKERR_ACCESS:
    return EACCESS;

  case SOCKERR_INTERNAL:
    return EIO;

  case SOCKERR_FD_INUSE:
    return EBADF;

  case SOCKERR_INFINITE_WAIT:
    return EINVAL;

  case SOCKERR_NOT_CONNECTED:
    return ENOTCONN;

  case SOCKERR_WOULD_BLOCK:
    return EWOULDBLOCK;

  case SOCKERR_NOT_LISTENING:
    return EWOULDBLOCK;

  default:
    printf ("socket: unhandled error %d %x\n", res, res);
    return EIO;
  }
}

static long __tcpfds = 0;
static long __udpfds = 0;
static long __otherselectablefds = 0;

#define ASSERTUDP(fd) if (fd >= 0 && fd < 32 && !(__udpfds & (1 << fd))) printf ("ERROR(%s): fd %d not UDP\n", __FUNCTION__);
#define ASSERTTCP(fd) if (fd >= 0 && fd < 32 && !(__tcpfds & (1 << fd))) printf ("ERROR(%s): fd %d not TCP\n", __FUNCTION__);
#define ASSERTSOCKET(fd) if (fd >= 0 && fd < 32 && !((__udpfds | __tcpfds) & (1 << fd))) printf ("ERROR(%s): fd %d not socket\n", __FUNCTION__);


int
socket_handler (__FSEXT_Fnumber func, int *rv, va_list ap) {
  int res, fd;
  void *buffer;
  size_t count;

  /*  printf ("----- socket_handler called ------\n"); */
  switch (func) {
  case __FSEXT_read:
  case __FSEXT_write:
    fd = va_arg(ap, int);
    buffer = va_arg(ap, void *);
    count = va_arg(ap, size_t);
    va_end(ap);

    /*    printf ("func %d: fd %d buf %x count %d\n", func, fd, buffer, count); */
    ASSERTTCP(fd);

    if (func == __FSEXT_read)
      res = sock_receive (fd, buffer, count);
    else
      res = sock_send (fd, buffer, &count);
    if (res < 0) {
      errno = converttoerrno(res);
      *rv = -1;
      return 1;
    }
    *rv = res;
    return 1;

  case __FSEXT_close:
    fd = va_arg(ap, int);
    va_end(ap);
    /* printf ("---close socket %d\n", fd); */
    ASSERTSOCKET(fd);
    res = sock_close (fd);
    if (res < 0) {
      errno = converttoerrno(res);
      *rv = -1;
      return 1;
    }
    __udpfds &= ~(1<<fd);
    __tcpfds &= ~(1<<fd);
    __FSEXT_set_function(fd, 0);
    _close(fd);
    *rv = 0;
    return 1;

  default:
    printf ("FSEXT handler--unknown func %d\n", func);
  }
  *rv = -1;
  return 1;
}

int
socket (int __family, int __type, int __protocol)
{
  int fd, proto, res;


  if (__family != AF_INET) {
    errno = EPROTONOSUPPORT;
    return -1;
  }

  if (__type == SOCK_STREAM && (!__protocol || (__protocol == IPPROTO_TCP))) {
    proto = 6;
    goto ok;
  }
  if (__type == SOCK_DGRAM && (!__protocol || (__protocol == IPPROTO_UDP))) {
    proto = 17;
    goto ok;
  }

  errno = EPROTONOSUPPORT;
  return -1;

ok:
  fd = __FSEXT_alloc_fd(socket_handler);
  if (fd == -1) {
    errno = ENFILE;
    return -1;
  }

  if (fd < 0 || fd > 31) {
    errno = EBADF;
    printf ("ERROR: socket(): fd out of range!\n");
    __FSEXT_set_function (fd, 0);
    _close(fd);
    return -1;
  }

  if (proto == 6)
    __tcpfds |= 1<<fd;
  else
    __udpfds |= 1<<fd;
  __file_handle_modes[fd] |= O_BINARY;
    
  res = sock_socket(proto, fd);
  if (res != 0) {
    errno = EACCESS;
    _close(fd);
    return -1;
  }
  /* printf ("---NEW socket %d (%s)\n", fd, proto == 6 ? "TCP" : "UDP"); */
  return fd;
}

int
bind (int __sockfd, const struct sockaddr *__my_addr,
		int __addrlen)
{
  int res;
  if (__addrlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_bind (__sockfd, 
		   ((struct sockaddr_in *) __my_addr)->sin_port,
		   ((struct sockaddr_in *) __my_addr)->sin_addr);
  if (res) {
    errno = converttoerrno(res);
    return -1;
  }
  return 0;
}

int
connect (int __sockfd, __const struct sockaddr *__serv_addr,
		int __addrlen)
{
  int res;

  if (__addrlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_connect(__sockfd, 
		     ((struct sockaddr_in *) __serv_addr)->sin_port,
		     ((struct sockaddr_in *) __serv_addr)->sin_addr.s_addr);
  if (res) {
    errno = converttoerrno(res);
    return -1;
  }
  return 0;
}

int
listen (int __sockfd, int __n)
{
  int res;
  res = sock_listen(__sockfd, __n);

  if (res) {
    errno = converttoerrno(res);
    return -1;
  }
  return 0;
}

int
accept (int __sockfd, __const struct sockaddr *__peer,
		int *__paddrlen)
{
  int res;
  int fd;

  fd = __FSEXT_alloc_fd(socket_handler);
  if (fd == -1) {
    errno = ENFILE;
    return -1;
  }

  if (*__paddrlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_accept(__sockfd, fd,
		    &((struct sockaddr_in *) __peer)->sin_port,
		    &((struct sockaddr_in *) __peer)->sin_addr.s_addr);
  if (res) {
    errno = converttoerrno(res);
    __FSEXT_set_function(fd, 0);
    _close(fd);
    return -1;
  }

  ((struct sockaddr_in *) __peer)->sin_family = AF_INET;
  *__paddrlen = sizeof(struct sockaddr_in);

  __tcpfds |= (1<<fd);
  __file_handle_modes[fd] |= O_BINARY;

  return fd;
}

int sendto (int __sockfd, __const void *__buff, size_t __len,
	    unsigned int __flags, __const struct sockaddr *__to,
	    int __tolen)
{
  int res;
  unsigned long len = __len;

  ASSERTUDP(__sockfd);

  if (__tolen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_sendto(__sockfd, 
		    ((struct sockaddr_in *) __to)->sin_port,
		    ((struct sockaddr_in *) __to)->sin_addr.s_addr,
		    __buff, len);
  if (!__quiet_socket)
      printf ("---sendto(%d) addr %08x:%04x len %d res %d\n", __sockfd, 
	      ntohl(((struct sockaddr_in *) __to)->sin_addr.s_addr),
	      ntohs(((struct sockaddr_in *) __to)->sin_port),
	      len, res);
  if (res < 0) {
    errno = converttoerrno(res);
    return -1;
  }

  return len;
}

int recvfrom (int __sockfd, void *__buff, size_t __len,
	      unsigned int __flags, struct sockaddr *__from,
	      int *__fromlen)
{
  int res;
  unsigned long len = __len;

  ASSERTUDP(__sockfd);

  if (*__fromlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_recvfrom(__sockfd, 
		      &((struct sockaddr_in *) __from)->sin_port,
		      &((struct sockaddr_in *) __from)->sin_addr.s_addr,
		      __buff, &len);
  if (res < 0) {
    errno = converttoerrno(res);
    return -1;
  }

  &((struct sockaddr_in *) __from)->sin_family = AF_INET;
  *__fromlen = sizeof(struct sockaddr_in);
  if (!__quiet_socket)
      printf (">>>recvfrom (%d) addr %08x:%04x len %d\n", __sockfd, 
	      ntohl(((struct sockaddr_in *) __from)->sin_addr.s_addr),
	      ntohs(((struct sockaddr_in *) __from)->sin_port),
	      len);
  return len;
}

void __addselectablefd (int fd)
{
  __otherselectablefds |= 1<<fd;
}

void __delselectablefd (int fd)
{
  __otherselectablefds &= ~(1<<fd);
}

int select (int numfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, 
	    struct timeval *timeout)
{
  int res;
  unsigned long timeo, mask;
  int i, n;

  if (numfds > 32) {
    printf ("ERROR: in select: numfds %d > 32\n", numfds);
    exit (1);
  }

  mask = 0;
  if (readfds)
    mask |= *(unsigned long *) readfds;
  if (writefds)
    mask |= *(unsigned long *) writefds;
  if (exceptfds)
    mask |= *(unsigned long *) exceptfds;
  if (mask & ~(__udpfds | __tcpfds | __otherselectablefds)) {
    printf ("ERROR: in select: selecting on non-socket FD\n");
    exit (1);
  }

  if (!timeout)
    timeo = 0xffffffff;
  else
    timeo = timeout->tv_sec*1000 + timeout->tv_usec/1000;

  res = sock_select(readfds, writefds, exceptfds, &timeo);
  if (res < 0) {
    errno = converttoerrno(res);
    return -1;
  }

  /* linux behavior wrt select() */
  if (timeout) {
    timeout->tv_sec = timeo/1000;
    timeout->tv_usec = 1000*(timeo % 1000);
  }

  mask = 0;
  if (readfds)
    mask |= *(unsigned long *) readfds;
  if (writefds)
    mask |= *(unsigned long *) writefds;
  if (exceptfds)
    mask |= *(unsigned long *) exceptfds;
  for (i = 0, n = 0; i < 32; i++)
    if (mask & (1<<i))
      n++;

  return n;
}

int getsockname (int __sockfd, struct sockaddr *__addr,
		 int *__paddrlen)
{
  int res;

  if (*__paddrlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_getname(__sockfd, 
		    &((struct sockaddr_in *) __addr)->sin_port,
		    &((struct sockaddr_in *) __addr)->sin_addr.s_addr);
  if (res) {
    errno = converttoerrno(res);
    return -1;
  }

  ((struct sockaddr_in *) __addr)->sin_family = AF_INET;
  *__paddrlen = sizeof(struct sockaddr_in);
  return 0;
}

int getpeername (int __sockfd, struct sockaddr *__addr,
		 int *__paddrlen)
{
  int res;

  if (*__paddrlen < sizeof(struct sockaddr_in)) {
    errno = EACCES;
    return -1;
  }

  res = sock_getpeer(__sockfd, 
		    &((struct sockaddr_in *) __addr)->sin_port,
		    &((struct sockaddr_in *) __addr)->sin_addr.s_addr);
  if (res) {
    errno = converttoerrno(res);
    return -1;
  }

  ((struct sockaddr_in *) __addr)->sin_family = AF_INET;
  *__paddrlen = sizeof(struct sockaddr_in);
  return 0;
}

#define GET_BLOCKING_MODE 0
#define SET_BLOCKING_MODE 1

int __djgpp_set_socket_blocking_mode (int sock, int onoff)
{
    int res;
    res = sock_getsetmode(sock, SET_BLOCKING_MODE, &onoff);
    if (res) {
	errno = converttoerrno(res);
	return -1;
    }
    return 0;
}


static struct servent services[] = {
{"rpc2portmap", NULL, 369, "udp"},
{"rpc2portmap", NULL, 369, "tcp"},
{"codaauth2", NULL, 370, "udp"},
{"codaauth2", NULL, 370, "tcp"},
{"venus", NULL, 2430, "udp"},
{"venus", NULL, 2430, "tcp"},
{"venus-se", NULL, 2431, "udp"},
{"venus-se", NULL, 2431, "tcp"},
{"codasrv", NULL, 2432, "udp"},
{"codasrv", NULL, 2432, "tcp"},
{"codasrv-se", NULL, 2433, "udp"},
{"codasrv-se", NULL, 2433, "tcp"},
{NULL, NULL, 0, NULL}};

static struct servent servreturn;

struct servent *
getservbyname(const char *name, const char *proto)
{
  struct servent *p;
  for (p = services; p->s_name; p++) {
    if (!strcmp(name, p->s_name) && 
	(!proto || !strcmp(proto, p->s_proto))) {
      servreturn = *p;
      servreturn.s_port = htons(servreturn.s_port);
      /*  printf ("getservbyname %s %s: %d\n", name, (proto ? proto : "ANY"), 
	      ntohs(servreturn.s_port)); */
      return &servreturn;
    }
  }
  printf ("getservbyname %s %s: not found\n", name, (proto ? proto : "ANY"));
  exit (1);
}


static void
strtolower (char *p)
{
    while (*p) {
	if (isupper(*p))
	    *p = tolower(*p);
	p++;
    }
}

static struct hostent host;
static unsigned long hostaddr;
static unsigned long *hostaddrs[] = {&hostaddr, 0};

#define HOSTS_FILE_NAME "C:/WINDOWS/HOSTS"

struct hostent *
gethostbyname(const char *name)
{
    FILE *fp = NULL;
    char line[200], *p, *namebuf = NULL, *lname;
    unsigned int a,b,c,d;
    /* printf ("gethostbyname %s\n", name); */
  
    host.h_name = name;
    host.h_aliases = 0;
    host.h_addrtype = AF_INET;
    host.h_length = 4;
    host.h_addr_list = (char **) &hostaddrs;

    if (sscanf (name, "%d.%d.%d.%d", &a, &b, &c, &d) == 4) {
	hostaddr = htonl((a << 24) | (b << 16) | (c << 8) | d);
	return &host;
    }

    /* otherwise, look in hosts file */
    fp = fopen(HOSTS_FILE_NAME, "r");
    if (!fp)
	goto errexit;

    /* canonicalize requested name -- strip leading
       and trailing whitespace and lowercase */
    namebuf = strdup(name);
    if (!namebuf) goto errexit;
    lname = strtok(namebuf, " \t\r\n");
    if (!lname) goto errexit;
    strtolower (lname);
 
    /* scan hosts file */
    while (fgets(line, sizeof(line), fp)) {
	p = strchr(line, '#');
	if (p) *p = '\0';
	p = strtok (line, " \t\r\n");
	if (!p) continue;
	if (sscanf (line, "%d.%d.%d.%d", &a, &b, &c, &d) != 4)
	    continue;
	hostaddr = htonl((a << 24) | (b << 16) | (c << 8) | d);
	while (p = strtok (NULL, " \t\r\n")) {
	    strtolower(p);
	    if (strcmp(p, lname) == 0) {
		hostaddr = htonl((a << 24) | (b << 16) | (c << 8) | d);
		free (namebuf);
		fclose (fp);
		return &host;
	    }
	}
    }

errexit:
    if (namebuf) free (namebuf);
    if (fp) fclose (fp);
    return NULL;
}

static char hostbuf[100];

struct hostent *
gethostbyaddr (const char *paddr, int len, int proto) 
{
  const unsigned char *addr = paddr;
  sprintf (hostbuf, "%d.%d.%d.%d", addr[0], addr[1], addr[2], addr[3]);
  host.h_name = hostbuf;
  host.h_aliases = 0;
  host.h_addrtype = AF_INET;
  return &host;
}
  
unsigned long
__djgpp_get_my_host()
{
    unsigned int a,b,c,d;
    static unsigned long my_host = 0;
    char buf[100];
    FILE *fp;

    if (my_host)
	return my_host;

    fp = fopen("myhost", "r");
    if (!fp) {
	printf ("ERROR: Could not open myhost file!\n");
	exit (1);
    }

    fgets(buf, sizeof(buf), fp);
    fclose(fp);

    if (sscanf (buf, "%d.%d.%d.%d", &a, &b, &c, &d) != 4) {
	printf ("ERROR: 'myhost' file entry (\"%s\") no good\n", buf);
	exit (1);
    }
    my_host = htonl((a << 24) | (b << 16) | (c << 8) | d);
    return my_host;
}

void
__djgpp_set_quiet_socket(int value)
{
   __quiet_socket = value & 1;
}
